/***************************************************************************
 ** $Id: basemusicnote.cpp,v 1.6 2004/05/13 02:57:20 wiecko Exp $
                          klearnnotes2
                          basemusicnote.cpp -  description
                             -------------------
    begin                : Tue Mar  2 02:41:00 CET 2004
    copyright            : (C) 2004 by Marek Wieckowski
    email                : wiecko AT users.sourceforge.net
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <ctype.h> 	  //tolower etc.

#include "globals.h"
#include "basemusicnote.h"
#include "scalenotation.h"
#include "oneletternotation.h"

namespace
{
    
//! Function to deallocate the notation global instance at exit.
void _destroyNotation();

//! This class takes care of the Notation Singleton
/*! At any given point in time, a specific notation is set as the current notation.
    This class is used to store that global instance, make sure there's always at least
    a default value for it and to make sure it's destructed at program's exit.

    The class uses the _destroyNotation helper function to schedule the destruction of the
    notation global instance when the program finishes, by registering it with atexit().
*/
class NotationKeeper
{
public:
   //! set the current notation
    static void setNotation(ScaleNotation* newNotation)
    {
       delete notation;
       notation = newNotation;
       
       if(!destructionHasBeenScheduled)
       {
          atexit(_destroyNotation);
          destructionHasBeenScheduled = true;
       }
    }
    
    //! retrieve the current notation
    /*! If no notation has set as current yet, it defaults to EnglishNotation. 
     */
    static ScaleNotation* getNotation()
    {
       if(notation == NULL)
       {
          // Notation not set yet; default to English Notation
          setNotation(new EnglishNotation);
       }
       return notation;
    }

private:
   NotationKeeper(); // disable instantiation
        
   // the global, current notation
   static ScaleNotation* notation; 

   // flag to indicate if atexit has already been called
   static bool destructionHasBeenScheduled;
   
   friend void _destroyNotation();
};

ScaleNotation* NotationKeeper::notation = NULL;
bool NotationKeeper::destructionHasBeenScheduled = false;

void _destroyNotation()
{
   delete NotationKeeper::notation;
   NotationKeeper::notation = NULL;
}

}

//================================================================================

BaseMusicNote::BaseMusicNote(){
  basePitch=0;
  accidental=0;
  return;
}

BaseMusicNote::BaseMusicNote(int initialPitch){
  basePitch=whiteOf(initialPitch);
  accidental=accidentalOf(initialPitch);
}

BaseMusicNote::BaseMusicNote(int initialBasePitch,int initialAccidental){
  basePitch=initialBasePitch;
  accidental=initialAccidental;
}

BaseMusicNote::BaseMusicNote(BaseMusicNote* copyFrom){
  basePitch=copyFrom->getBasePitch();
  accidental=copyFrom->getAccidental();
}



int BaseMusicNote::numOfNoteInOctave(int pitch){  return(pitch%12);}
bool BaseMusicNote::isWhite(int pitch)
{
  switch(numOfNoteInOctave(pitch))
    {
    case 1:
    case 3:
    case 6:
    case 8:
    case 10:
      return false;
      //      break;
    default:
      return true;
    }
}
int BaseMusicNote::whiteOf(int pitch)
{
  if(isWhite(pitch))
    return pitch;
  else
    return pitch-1;
}
int BaseMusicNote::accidentalOf(int pitch)
{
  if(isWhite(pitch))
    return 0;
  else
    return 1;
}

QString BaseMusicNote::stringName(int numInOctave, int accident, bool shorty)
{
  if ((accident>1 )||(accident<-1))
    return stringName(pitch(numInOctave,accident),shorty);
  // just a simple assertion; for example if someone calls:
  // stringName(3,2) (meaning D double sharp)
  // this will return: pitch=3+2=5, and althogether "F" which is better
  // than nothing

  ScaleNotation::ScaleNote note = ScaleNotation::evC;
  switch(numInOctave){
     case 0:
       note = ScaleNotation::evC;
       break;
       
     case 2:
       note = ScaleNotation::evD;
       break;
       
     case 4:
       note = ScaleNotation::evE;
       break;
       
     case 5:
       note = ScaleNotation::evF;
       break;
       
     case 7:
       note = ScaleNotation::evG;
       break;
       
     case 9:
       note = ScaleNotation::evA;
       break;
       
     case 11:
       note = ScaleNotation::evB;
       break;

     default:
       // unmatched (e.g non-white or too big/too small) numInOctave
       return stringName(pitch(numInOctave,accident),shorty);      
  }
  
  ScaleNotation::NoteAccident noteAccident = ScaleNotation::evNatural;
  switch(accident)
  {
    case 0: 
        noteAccident = ScaleNotation::evNatural;
        break;
        
    case 1: 
        noteAccident = ScaleNotation::evSharp;
        break;
        
    case -1:
        noteAccident = ScaleNotation::evFlat;
        break;
  }

  if (shorty)
    return getNotation().getNoteLabel(note, noteAccident);

  return getNotation().getNoteLongLabel(note, noteAccident);
  
  // unmatched (e.g non-white or too big/too small) numInOctave
  // is handled in 'default' of each switch:
  /*switch(notation){
  case notationEnglish:
    switch(accident){
    case 0://natural:
      switch(numInOctave){
      case 0:
	return QString(tr("C","name of a note"));
      case 2:
	return QString(tr("D","name of a note"));
      case 4:
	return QString(tr("E","name of a note"));
      case 5:
	return QString(tr("F","name of a note"));
      case 7:
	return QString(tr("G","name of a note"));
      case 9:
	return QString(tr("A","name of a note"));
      case 11:
	return QString(tr("B","name of a note"));
      default:
	return stringName(pitch(numInOctave,accident));
      }
    case 1://sharp:
      switch(numInOctave){
      case 0:
	return QString(tr("C %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 2:
	return QString(tr("D %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 4:
	return QString(tr("E %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 5:
	return QString(tr("F %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 7:
	return QString(tr("G %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 9:
	return QString(tr("A %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 11:
	return QString(tr("B %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      default:
	return stringName(pitch(numInOctave,accident));
      }
    case -1://flat
      switch(numInOctave){
      case 0:
	return QString(tr("C %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 2:
	return QString(tr("D %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 4:
	return QString(tr("E %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 5:
	return QString(tr("F %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 7:
	return QString(tr("G %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 9:
	return QString(tr("A %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 11:
	return QString(tr("B %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      default:
	return stringName(pitch(numInOctave,accident));
      }
    }
  case notationGerman:
    switch(accident){
    case 0://natural:
      switch(numInOctave){
      case 0:
	return QString(tr("C","name of a note"));
      case 2:
	return QString(tr("D","name of a note"));
      case 4:
	return QString(tr("E","name of a note"));
      case 5:
	return QString(tr("F","name of a note"));
      case 7:
	return QString(tr("G","name of a note"));
      case 9:
	return QString(tr("A","name of a note"));
      case 11:
	return QString(tr("H","name of a note"));
      default:
	return stringName(pitch(numInOctave,accident));
      }
    case 1://sharp:
      switch(numInOctave){
      case 0:
	return QString(tr("C %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 2:
	return QString(tr("D %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 4:
	return QString(tr("E %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 5:
	return QString(tr("F %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 7:
	return QString(tr("G %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 9:
	return QString(tr("A %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      case 11:
	return QString(tr("H %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("sharp")));
      default:
	return stringName(pitch(numInOctave,accident));
      }
    case -1://flat
      switch(numInOctave){
      case 0:
	return QString(tr("C %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 2:
	return QString(tr("D %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 4:
	return QString(tr("E %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 5:
	return QString(tr("F %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 7:
	return QString(tr("G %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 9:
	return QString(tr("A %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      case 11:
	return QString(tr("H %1","name of a note; #1 is 'sharp' or 'flat'").
		       arg(tr("flat")));
      default:
	return stringName(pitch(numInOctave,accident));
      }
    }
  }
 // we should never ever be here, but just to get rid of warnings:
  return stringName(pitch(numInOctave,accident));*/
}

QString BaseMusicNote::stringName(int pitch, bool shorty)
{
  return stringName(numOfNoteInOctave(whiteOf(pitch)),accidentalOf(pitch),
		    shorty);
}


int BaseMusicNote::pitch(int basePtch,int accid){return basePtch+accid;}

int BaseMusicNote::whiteNumberToTotalNumberInOctave(int number)
{
  while(number<0)
    number +=8;
  number=number%8;
  number*=2;
  if(number>4)
    number--;
  return(number);
}

int BaseMusicNote::pitchToWhiteNumberInOctave(int pitch)
{
  pitch=pitch%12;
  if(pitch>4)
    pitch++;
  return(pitch/2);
}


int BaseMusicNote::pitch(){return pitch(basePitch,accidental);}

int BaseMusicNote::numOfNoteInOctave(){return numOfNoteInOctave(pitch());}

QString BaseMusicNote::stringName(bool shorty)
{
  return stringName(numOfNoteInOctave(basePitch),accidental,shorty);
}

bool BaseMusicNote::isWhite(){return isWhite(pitch());}
int BaseMusicNote::getBasePitch(){return basePitch;}
int BaseMusicNote::getAccidental(){return accidental;}

void BaseMusicNote::setBasePitch(int newBasePitch){basePitch=newBasePitch;}
void BaseMusicNote::setAccidental(int newAccidental){accidental=newAccidental;}

//! Returns the current notation
/*! If the current notation hasn't been set yet (by calling setNotation)
    the EnglishNotation will be set as the default notation and returned to the
    caller.
*/
const ScaleNotation& BaseMusicNote::getNotation()
{
    return *NotationKeeper::getNotation();
}

//! Sets the current notation to newNotation
/*! Note that this function will take ownership of the given notation object
    and take care of its destruction. Therefore, it has to be a new dynamically allocated
    object.
*/
void BaseMusicNote::setNotation(ScaleNotation* newNotation)
{
    NotationKeeper::setNotation(newNotation);
}

BaseMusicNote* BaseMusicNote::noteMatchingKeyEvent(QKeyEvent *k)
{
  char key = toupper(k->ascii());
  DEBUG_VAR(QChar(key));
  for(int i=0;i<7;i++)
    if(key == getNotation().getNoteChar(static_cast<ScaleNotation::ScaleNote>(6-i)))
      {
	debuginfo("Found:");
	DEBUG_VAR(whiteNumberToTotalNumberInOctave(i));
	return new BaseMusicNote(whiteNumberToTotalNumberInOctave(i));
      }
  debuginfo("Key: note not found!");
  return 0;
}

bool BaseMusicNote::isEqualName(BaseMusicNote* note1,
				BaseMusicNote* note2)
{
  return (note1->stringName()==note2->stringName());
}
