/***************************************************************************
 ** $Id: basemusicnote.h,v 1.6 2004/05/13 02:57:20 wiecko Exp $
                          klearnnotes2
                          basemusicnote.cpp -  description
                             -------------------
    begin                : Tue Mar  2 02:41:00 CET 2004
    copyright            : (C) 2004 by Marek Wieckowski
    email                : wiecko AT users.sourceforge.net
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KLN_BASEMUSICNOTE_H
#define KLN_BASEMUSICNOTE_H

#include <qstring.h>
#include <qobject.h>

class ScaleNotation;

//! a class holding note to pitch methods
/*! Note: in all the following I call midi tune number a "pitch". */
class BaseMusicNote : public QObject
{
  Q_OBJECT
    // Q_OBJECT needed for tr() to work;
 public:
  BaseMusicNote();
  BaseMusicNote(int initialPitch);
  BaseMusicNote(int initialBasePitch,int initialAccidental);
  BaseMusicNote(BaseMusicNote* copyFrom);

  //! the pitch of middle C (=60)
  static const int pitchOfMiddleC = 60;
  //!  the pitch of A440 (=69)
  static const int pitchOfA440 = 69;
 
  //! number of the note in it's octave; C=0, C#/Db=1, D=2 and so on
  static int numOfNoteInOctave(int pitch);
  //! name of the note in QString form, dependent on notation and translations
  static QString stringName(int integerName, int accident, bool shorty=false);
  //! name of the note in QString form, dependent on notation and translations
  static QString stringName(int pitch, bool shorty=false);

  //! true if the pitch represents a white key
  /*! This means, the pitch can be represented as a note without any
    accidentals. */
  static bool isWhite(int pitch);
  //! non-accidented note; no change or pitch-1
  /*! All pitches have the basic, simplest representation. For all "white
    keys" this is simply the pitch of the key. For "black keys" it will be
    pitch-1.

    This means, the whiteOf(61)==60 [note: 61 is C#/Db and 60 is C],
    and whiteOf(60)==60

    Together whiteOf() and accidentalOf() split a pitch into base note 
    + accidental pair.
  */
  static int whiteOf(int pitch);
  //! the basic accidental for the pitch: 0(natural) or 1(sharp)
  /*! All pitches have the basic, simplest representation. For all "white
    keys" this is simply the pitch of the key, and this function will return 0.
    For "black keys" the base is pitch-1, and this function returns +1.

    This means, the accidentalOf(61)==1, and accidentalOf(60)==0.
  */
  static int accidentalOf(int pitch);

  //!returns the sum of the arguments
  /*! accid shoud be -2 for double flat, -1 for flat, 0 for natural ,
    1 for sharp and 2 for double sharp */
  static int pitch(int basePtch,int accid);

  //! takes number of a base name and adds number of black keys "missed"
  /*! This method converts number of a white key when counting only white
    keys to its number when all keys are counted. Therefore, the argument
    whiteNum is 0 for C, 1 for D, 2 for E, 3 for F, 4 for G... (it can
    be bigger than 8, and for negative arguments full 8s are added until
    it is positive).

    The returned values are: 0 for C, 2 for D, 4 for E, 5 for F, 7 for G...
    and the maximum is 11 for B.
  */
  static int whiteNumberToTotalNumberInOctave(int whiteNum);

  //! reverse of whiteNumberToTotalNumberInOctave()
  /*! It is called pitchTo... because argument is not limited to the range 
     0...11 ; for other values it is moduled to this range. 
     
     For pitches which do not correspond to white buttons, nubmer of 
     half-tone lower white button is retured. Therefore for any C and C#
     pitch the result is 0, for D and D# 1, for E 2, for F and F# 3 and so on.
  */
  static int pitchToWhiteNumberInOctave(int pitch);

  //! tries to convert QKeyEvent to an example note
  /*! e.g. if one uses English notes naming and pressess 'C' this returns
    C0 note (of pitch 60); if unsuccessful returns 0 */
  static BaseMusicNote* noteMatchingKeyEvent(QKeyEvent *k);

  //! returns true if names of both notes match
  static bool isEqualName(BaseMusicNote* note1,
			  BaseMusicNote* note2);

  /////////////////////////
  // non-static members working on BaseMusicNote's properties
  /////////////////////////

  //! returns pitch( basePitch, accidental)
  virtual int pitch();
  //! returns numOfNoteInOctave(pitch())
  virtual int numOfNoteInOctave();
  //! stringName(basePitch, accidental)
  virtual QString stringName(bool shorty=false);
  //! true if the note's pitch represents a white key
  virtual bool isWhite();

  virtual int getBasePitch();
  virtual int getAccidental();

  virtual void setBasePitch(int);
  virtual void setAccidental(int);

  static const ScaleNotation& getNotation();
  static void setNotation(ScaleNotation* newNotation);

 protected:
  
  int basePitch;
  int accidental;

};


#endif       // KLN_BASEMUSICNOTE_H
