/***************************************************************************
 ** $Id: randomobject.h,v 1.6 2004/04/27 02:28:22 wiecko Exp $
                          klearnnotes2
                          randomobject.h -  description
                             -------------------
    begin                : Sat Nov 29 01:54:17 CET 2003
    copyright            : (C) 2003 by Marek Wieckowski
    email                : wiecko AT users.sourceforge.net
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KLN_RANDOMOBJECT
#define KLN_RANDOMOBJECT

#include"basemusicnote.h"
#include"exercise.h"
#include<qptrlist.h>
#include<qframe.h>

//! A parent class for all randomly-chosen objects.
/*! Random objects are those, which can be randomly chosen as a test
	question: notes in the staff, positions on a fretboard etc.

  This is inherited parent class which defines all random object specific
  properties.
*/
class RandomObject
{
 public:
  RandomObject(Exercise* myExercise_ = 0); //the creator
  virtual ~RandomObject(); //the destructor

  //! resets statistics data to pre-test defaults
  void resetStats();

  //! adds one correct answer to statOK and increases statTIME by t
  void addCorrect(int t);
  //! adds one wrong answer to statWR and increases statTIME by t
  void addWrong(  int t);

  //! returns number of correct answers, statOK
  int getCorrect();
  //! returns number of incorrect answers,  statWR
  int getWrong();
  //! returns time spent on all questions about this randomObject, statTIME
  int getTime();
  //! returns speed of answers about this randomObject, statSPEED
  double getSpeed();
 //! returns re-asking probability weight,  statPROBAB
  double getProbab();

  //! run whenever setupActiveNotes is run; this runs Exercise::isActiveXXX()
  virtual bool checkIfIsActive();
  //! doesn't check anything, just returns 'active'
  virtual bool isActiveForTest();

  //! used to highlight the RandomObject when it was selected as a question
  virtual void showForTest();
  //! used to remove highlight done by showForTest() and in the beginning of a test
  virtual void hideForTest();

 protected:
  //! Calculates statSPEED and statPROBAB
  /*! Called by addCorrect() and addWrong(), that is, whenever an 
	answer is given. */
  void updateStats(); 

  //! number of correct answers
  int statOK;
  //! number of wrong answers
  int statWR;
  //! time spent on questions about this randomObject
  int statTIME;
  //! speed of answers about this randomObject
  double statSPEED;

  //! a probability weight
  /*! a weight determining 
      probability of re-asking question about this object;
    * -  increased by wrong or slow answers;
    * -  decreased by right and fast answers;

	See "how does it work?" section of helpfiles for details.
	*/
  double statPROBAB;  

  //! used for checking if this RandomObject should be considered active
  Exercise* myExercise;
  //! true if it is active for test
  bool active;
};

//! OBSOLETE! Initializes suggestion[] array
/*! OBSOLETE!
	* - suggestion[] is an empty array of force-scheduled questions
	* - maxSug is size of suggestions[]
	* - questions[] is an array with all possible questions for this 
	    test
	* - numObjects is size of questions[]

	It schedules all questions[] into a first forced run of suggestion[].

	In future should be redefined as a static member of RandomObject class.

	See "how does it work?" section of helpfiles for details.
	*/
void RandomObject_init(RandomObject* suggestion[], const int maxSug,
		       RandomObject* questions[], int numObjects);


//! OBSOLETE! Draws (randomly chooses) next question
/*! OBSOLETE!
	* - quenum is current question number
	* - lastObject[] is one-element array which holds previous question
	* - newObject[] is one element array, where new chosen question is returned.

	Other arguments like in RandomObject_init().

	This function checks if there is a question scheduled in suggestion[]
	for current question number. If yes, newObject[0] is set to this force-
	scheduled question and function exits. If not, one of questions[] is 
	randomly chosen (questions with greater statPROBAB are more probable),
	and newObject[0] is set to this randomly drawn object.

	In future should be redefined as a static member of RandomObject class.

	See "how does it work?" section of helpfiles for details.
	*/
void RandomObject_next(int quenum, RandomObject* lastObject[],
		       RandomObject* newObject[],
		       RandomObject* suggestion[],
		       const int maxSug,
		       RandomObject* questions[], int numObjects);

//! OBSOLETE! Force-schedules a question for future.
/*! OBSOLETE!
	* - quenum is current question number
	* - object[] is one-element array which holds the question 
	the user made mistake about.

	Other arguments like in RandomObject_init().

	This function is used whenever user makes a mistake. It adds 
	object[0] as a force-scheduled questions (i.e. into suggestion[])
	for 2nd, 5th and 8th next question or, if there are already some
	question scheduled for these, for next available.

	In future should be redefined as a static member of RandomObject class.

	See "how does it work?" section of helpfiles for details.
	*/
void RandomObject_schedule(int quenum,RandomObject* object[],
			   RandomObject* suggestion[], const int maxSug);


//! class holding and manipulating all the quesitions
/*! All wigets (like box of namebuttons, staff with notes, fretboard with
pieces of fret) should inherit this to force common interface to 
starting/stopping exercise, changing available questions pool etc. 

Therefore, this is a common interface to:
* - ask a question
* - provide an answer
* - check if an answer is right
* - react on right/wrong answer (highlight RandomObjects)
*/
class RandomObjectsContainer :public QFrame //: public QObject
{
  Q_OBJECT
 public:
  RandomObjectsContainer(QWidget* parent=0);
  virtual ~RandomObjectsContainer();

  //! says if the note should be considered active by other wigets
  /*! should be run by Exercise when any widget asks it if it should be
    active */
  virtual bool isActiveNote(BaseMusicNote*);
  virtual bool isActiveName(BaseMusicNote*);
  virtual bool isActivePitch(BaseMusicNote*);

  public slots:
    //! checks if provided answer matches the question
    /*! This checks exact note; can be used if user answered by clicking on
      a staff.

    It emits one signal answerWasRight().*/
    virtual bool checkAnswerByNote(BaseMusicNote*);
    //! checks if provided answer matches the question
    /*! This checks name; can be used if user answered by clicking on
      a namebutton or pressed a comp. keyboard key. 

      It emits one signal answerWasRight().*/
    virtual bool checkAnswerByName(BaseMusicNote*);
    //! checks if provided answer matches the question
    /*! This checks pitch; can be used if user answered with her instrument.


    It emits one signal answerWasRight().       */
    virtual bool checkAnswerByPitch(BaseMusicNote*);

    //! if test is not started - starts it and chooses the first question
    virtual void startTest(int length=60);
    //! if test is running - stops it and emits stopped() to trigger summary
    virtual void stopTest();

    //! highlights the answer object to right(green) or wrong(red)
    /*! connected to other widget's     answerWasRight() */
    virtual void highlightRight(BaseMusicNote*,bool isItRight=true);

    //! setup which questions should be active
    /*! run whenever any widget changes range/kind of notes which should be
      considered active */
    virtual void setupActiveQuestions();

    //! used for growing timeSpent, advance animation etc.
    /*! slot activated by clock ticking */
    virtual void advance();

    //! during a test, returns the RandomObject selected as a question
    virtual RandomObject* getCurrentQuestion();

    //! starts a test
    virtual void initTest();
    //! draws a new question
    virtual void nextQuestion();
    //! after a wrong answer force-schedules the current qustion for future
    virtual void schedule();

 signals:
    // signal returning true if the answer was right
    /*! or false otherwise. Emitted by checkAnswerXXX().

    Example for name to note exercise:
    * - qusetion=one of pushButtons in buttonBox;
    * - user clicks on a note in the staff; this note is provieded as an answer
        to buttonBox's checkAnswerByName;
    * - if the answer was right the signal
    answerWasRight(note,true) is emitted; otherwise answerWasRight(note,false) 
    is emitted
    * - this signal can be connected back to the canvas so that the clicked
    note knew if it should get highlighted in red (wrong) or green (right).
*/
    void answerWasRight(BaseMusicNote*,int timeSpentOnThis,
			bool wasItRight=true);
    //! emmited when stopping a test (is it really used????)
    void stopped();
    //! signals emitted when user answers
    /*! should be connected to other wiget's checkanswer() */
    void answerByNote(BaseMusicNote*);
    void answerByName(BaseMusicNote*);
    void answerByPitch(BaseMusicNote*);

    void clockTick();
    //! emitted at the beginning of any test to reset visible results display
    void resetStatistics();
    //! emitted to output a string to main window's comments box
    void comment(QString);

 protected:
    void  debugShowSuggestions(int activequenum);
    QPtrList<RandomObject> activeQuestions;
    QPtrList<RandomObject> allQuestions;
    //! not used - I guess it will be removed
    static QPtrList<RandomObjectsContainer> allContainers;
    //! array of force-scheduled questions
    RandomObject* suggestion[defMAXseria];
    //    const int maxSug=defMAXseria;
    //! current question number
    int questionNumber;
    //! current question
    RandomObject* currentQuestion;
    //! number of questions in one test
    int testLength;
    QTimer* testClock;
    int timeSpent;
}; 
#endif
